@@ -63,6 +63,8 @@
 
 #include "socket.h"
 
+#include "mongoose.h"
+
 #include "timeops.h"
 
 #ifdef WIN32
@@ -80,6 +82,18 @@
 
 using namespace std;
 
+//------------------------------------------------------------------------------
+// Mongoose web server
+struct mg_server *server = (mg_server *)NULL;
+void *poll_server(void *);
+void start_web_server();
+void close_server();
+void get_next_port_number();
+
+int flmsg_webserver_portnbr = 8080;
+char flmsg_webserver_szportnbr[5];
+//------------------------------------------------------------------------------
+
 const char *options[] = {\
 "flmsg unique options",
 "--help",
@@ -152,6 +166,7 @@ string ICS_dir = "";
 string ICS_msg_dir = "";
 string ICS_tmp_dir = "";
 string CSV_dir = "";
+string CUSTOM_dir = "";
 string XFR_dir = "";
 string FLMSG_temp_dir = "";
 string FLMSG_custom_dir = "";
@@ -715,6 +730,11 @@ void extract_text(string &buffer, const char *fname)
 		read_csvbuffer(buffer);
 		if (fname) def_csv_filename = fname;
 		select_form(selected_form);
+	} else if (buffer.find("<customform>") != string::npos) {
+		selected_form = CUSTOM;
+		read_custombuffer(buffer);
+		if (fname) def_custom_filename = fname;
+		select_form(selected_form);
 	} else if (!exit_after_print)
 		fl_alert2(_("Not an flmsg data file"));
 }
@@ -775,6 +795,7 @@ int eval_transfer_size()
 		case PLAINTEXT:	return eval_pt_fsize();
 		case BLANK:		return eval_blank_fsize();
 		case CSV:		return eval_csv_fsize();
+		case CUSTOM:	return eval_custom_fsize();
 		case TRANSFER:	return eval_transfer_fsize();
 		case MARSDAILY:	return eval_mars_daily_fsize();
 		case MARSINEEI:	return eval_mars_ineei_fsize();
@@ -816,6 +837,7 @@ void cb_new()
 		case BLANK: cb_blank_new(); break;
 		case TRANSFER: cb_transfer_new(); break;
 		case CSV: cb_csv_new(); break;
+		case CUSTOM: cb_custom_new(); break;
 		case MARSDAILY: cb_mars_daily_new(); break;
 		case MARSINEEI: cb_mars_ineei_new(); break;
 		case MARSNET: cb_mars_net_new(); break;
@@ -855,6 +877,7 @@ void cb_import()
 		case PLAINTEXT:
 		case BLANK:
 		case CSV:
+		case CUSTOM:
 		case REDXSNW:
 		case REDX5739:
 		case REDX5739A:
@@ -895,6 +918,7 @@ void cb_export()
 		case PLAINTEXT:
 		case BLANK:
 		case CSV:
+		case CUSTOM:
 		case REDXSNW:
 		case REDX5739:
 		case REDX5739A:
@@ -985,6 +1009,9 @@ void wrap_import(const char *fname)
 			} else if (inpbuffer.find("<csvform>") != string::npos) {
 				selected_form = CSV;
 				cb_csv_wrap_import(filename, inpbuffer);
+			} else if (inpbuffer.find("<customform>") != string::npos) {
+				selected_form = CUSTOM;
+				cb_custom_wrap_import(filename, inpbuffer);
 			} else if (inpbuffer.find("<mars_daily>") != string::npos) {
 				selected_form = MARSDAILY;
 				cb_mars_daily_wrap_import(filename, inpbuffer);
@@ -1117,6 +1144,7 @@ void cb_wrap_export()
 		case PLAINTEXT: cb_pt_wrap_export(); break;
 		case BLANK: cb_blank_wrap_export(); break;
 		case CSV: cb_csv_wrap_export(); break;
+		case CUSTOM: cb_custom_wrap_export(); break;
 		case TRANSFER: cb_transfer_wrap_export(); break;
 		case MARSDAILY: cb_mars_daily_wrap_export(); break;
 		case MARSINEEI: cb_mars_ineei_wrap_export(); break;
@@ -1167,6 +1195,7 @@ void cb_wrap_autosend()
 		case PLAINTEXT: cb_pt_wrap_autosend(); break;
 		case BLANK: cb_blank_wrap_autosend(); break;
 		case CSV: cb_csv_wrap_autosend(); break;
+		case CUSTOM: cb_custom_wrap_autosend(); break;
 		case TRANSFER: cb_transfer_wrap_autosend(); break;
 		case MARSDAILY: cb_mars_daily_wrap_autosend(); break;
 		case MARSINEEI: cb_mars_ineei_wrap_autosend(); break;
@@ -1206,6 +1235,7 @@ void cb_load_template()
 		case PLAINTEXT: cb_pt_load_template(); break;
 		case BLANK: cb_blank_load_template(); break;
 		case CSV: cb_csv_load_template(); break;
+		case CUSTOM: cb_custom_load_template(); break;
 		case MARSDAILY: cb_mars_daily_load_template(); break;
 		case MARSINEEI: cb_mars_ineei_load_template(); break;
 		case MARSNET: cb_mars_net_load_template(); break;
@@ -1245,6 +1275,7 @@ void cb_save_template()
 		case PLAINTEXT: cb_pt_save_template(); break;
 		case BLANK: cb_blank_save_template(); break;
 		case CSV: cb_csv_save_template(); break;
+		case CUSTOM: cb_custom_save_template(); break;
 		case MARSDAILY: cb_mars_daily_save_template(); break;
 		case MARSINEEI: cb_mars_ineei_save_template(); break;
 		case MARSNET: cb_mars_net_save_template(); break;
@@ -1283,6 +1314,7 @@ void cb_save_as_template()
 		case PLAINTEXT: cb_pt_save_as_template(); break;
 		case BLANK: cb_blank_save_as_template(); break;
 		case CSV: cb_csv_save_as_template(); break;
+		case CUSTOM: cb_custom_save_as_template(); break;
 		case MARSDAILY: cb_mars_daily_save_as_template(); break;
 		case MARSINEEI: cb_mars_ineei_save_as_template(); break;
 		case MARSNET: cb_mars_net_save_as_template(); break;
@@ -1321,6 +1353,7 @@ void cb_open()
 		case PLAINTEXT: cb_pt_open(); break;
 		case BLANK: cb_blank_open(); break;
 		case CSV: cb_csv_open(); break;
+		case CUSTOM: cb_custom_open(); break;
 		case MARSDAILY: cb_mars_daily_open(); break;
 		case MARSINEEI: cb_mars_ineei_open(); break;
 		case MARSNET: cb_mars_net_open(); break;
@@ -1372,6 +1405,7 @@ void cb_save_as()
 		case REDX5739B: cb_redx_5739B_save_as(); break;
 		case BLANK: cb_blank_save_as(); break;
 		case CSV: cb_csv_save_as(); break;
+		case CUSTOM: cb_custom_save_as(); break;
 		default: ;
 	}
 }
@@ -1410,6 +1444,7 @@ void cb_save()
 		case REDX5739B: cb_redx_5739B_save(); break;
 		case BLANK: cb_blank_save(); break;
 		case CSV: cb_csv_save(); break;
+		case CUSTOM: cb_custom_save(); break;
 		default: ;
 	}
 }
@@ -1448,6 +1483,7 @@ void cb_html()
 		case REDX5739B: cb_redx_5739B_html(); break;
 		case BLANK: cb_blank_html(); break;
 		case CSV: cb_csv_html(); break;
+		case CUSTOM: cb_custom_html(); break;
 		default: ;
 	}
 }
@@ -1506,6 +1542,7 @@ void cb_text()
 		case REDX5739A: cb_redx_5739A_textout(); break;
 		case REDX5739B: cb_redx_5739B_textout(); break;
 		case CSV: cb_csv_textout(); break;
+		case CUSTOM: cb_custom_textout(); break;
 		case BLANK:
 		default: cb_blank_textout();
 	}
@@ -1523,6 +1560,7 @@ void cb_exit()
 	catch (...) {
 	}
 	close_xmlrpc();
+	close_server();
 	debug::stop();
 	exit(0);
 }
@@ -1640,6 +1678,9 @@ void show_filename(string p)
 		case CSV:
 			base_csv_filename = fl_filename_name(p.c_str());
 			break;
+		case CUSTOM:
+			base_custom_filename = fl_filename_name(p.c_str());
+			break;
 		case TRANSFER:
 			base_transfer_filename = fl_filename_name(p.c_str());
 			break;
@@ -1715,23 +1756,6 @@ void after_start(void *)
 {
 	check_mycall();
 
-	LOG_INFO("FLMSG_dir        %s", FLMSG_dir.c_str());
-	LOG_INFO("ARQ_dir          %s", ARQ_dir.c_str());
-	LOG_INFO("ARQ_files_dir    %s", ARQ_files_dir.c_str());
-	LOG_INFO("ARQ_recv_dir     %s", ARQ_recv_dir.c_str());
-	LOG_INFO("ARQ_send_dir     %s", ARQ_send_dir.c_str());
-	LOG_INFO("WRAP_dir         %s", WRAP_dir.c_str());
-	LOG_INFO("WRAP_recv_dir    %s", WRAP_recv_dir.c_str());
-	LOG_INFO("WRAP_send_dir    %s", WRAP_send_dir.c_str());
-	LOG_INFO("WRAP_auto_dir    %s", WRAP_auto_dir.c_str());
-	LOG_INFO("ICS_dir          %s", ICS_dir.c_str());
-	LOG_INFO("ICS_msg_dir      %s", ICS_msg_dir.c_str());
-	LOG_INFO("ICS_tmp_dir      %s", ICS_tmp_dir.c_str());
-	LOG_INFO("CSV_dir          %s", CSV_dir.c_str());
-	LOG_INFO("Transfer dir     %s", XFR_dir.c_str());
-	LOG_INFO("FLMSG_temp_dir   %s", FLMSG_temp_dir.c_str());
-	LOG_INFO("FLMSG_custom_dir %s", FLMSG_custom_dir.c_str());
-
 	def_203_filename = ICS_msg_dir;
 	def_203_filename.append("default"F203_EXT);
 	def_203_TemplateName = ICS_tmp_dir;
@@ -1796,13 +1820,11 @@ void after_start(void *)
 	def_rg_filename.append("default"RGFILE_EXT);
 	def_rg_TemplateName = ICS_tmp_dir;
 	def_rg_TemplateName.append("default"RGTEMP_EXT);
-	set_rg_choices();
 
 	iaru_def_filename = ICS_msg_dir;
 	iaru_def_filename.append("default"IARU_FILE_EXT);
 	iaru_def_template_name = ICS_tmp_dir;
 	iaru_def_template_name.append("default"IARU_TEMP_EXT);
-	iaru_set_choices();
 
 	cap105_def_filename = ICS_msg_dir;
 	cap105_def_filename.append("default"CAP105_FILE_EXT);
@@ -1830,6 +1852,11 @@ void after_start(void *)
 	def_csv_TemplateName = ICS_tmp_dir;
 	def_csv_TemplateName.append("default"CSVTEMP_EXT);
 
+	def_custom_filename = ICS_msg_dir;
+	def_custom_filename.append("default"CUSTOMFILE_EXT);
+	def_custom_TemplateName = ICS_tmp_dir;
+	def_custom_TemplateName.append("default"CUSTOMTEMP_EXT);
+
 	def_mars_daily_filename = ICS_msg_dir;
 	def_mars_daily_filename.append("default"FMARSDAILY_EXT);
 	def_mars_daily_TemplateName = ICS_tmp_dir;
@@ -1993,6 +2020,8 @@ int main(int argc, char *argv[])
 
 	mainwindow = flmsg_dialog();
 	mainwindow->callback(exit_main);
+	set_rg_choices();
+	iaru_set_choices();
 
 #if FLMSG_FLTK_API_MAJOR == 1 && FLMSG_FLTK_API_MINOR == 3
 	Fl::add_handler(default_handler);
@@ -2023,19 +2052,40 @@ int main(int argc, char *argv[])
 	FSEL::create();
 
 	checkdirectories();
+	load_custom_menu();
+	get_next_port_number();
 
 	string debug_file = FLMSG_dir;
-	debug_file.append("debug_log.txt");
+	debug_file.append("debug_log_").append(flmsg_webserver_szportnbr).append(".txt");
 	debug::start(debug_file.c_str());
 
+	LOG_INFO("FLMSG_dir        %s", FLMSG_dir.c_str());
+	LOG_INFO("ARQ_dir          %s", ARQ_dir.c_str());
+	LOG_INFO("ARQ_files_dir    %s", ARQ_files_dir.c_str());
+	LOG_INFO("ARQ_recv_dir     %s", ARQ_recv_dir.c_str());
+	LOG_INFO("ARQ_send_dir     %s", ARQ_send_dir.c_str());
+	LOG_INFO("WRAP_dir         %s", WRAP_dir.c_str());
+	LOG_INFO("WRAP_recv_dir    %s", WRAP_recv_dir.c_str());
+	LOG_INFO("WRAP_send_dir    %s", WRAP_send_dir.c_str());
+	LOG_INFO("WRAP_auto_dir    %s", WRAP_auto_dir.c_str());
+	LOG_INFO("ICS_dir          %s", ICS_dir.c_str());
+	LOG_INFO("ICS_msg_dir      %s", ICS_msg_dir.c_str());
+	LOG_INFO("ICS_tmp_dir      %s", ICS_tmp_dir.c_str());
+	LOG_INFO("CSV_dir          %s", CSV_dir.c_str());
+	LOG_INFO("CUSTOM_dir       %s", CUSTOM_dir.c_str());
+	LOG_INFO("Transfer dir     %s", XFR_dir.c_str());
+	LOG_INFO("FLMSG_temp_dir   %s", FLMSG_temp_dir.c_str());
+
 	LOG_INFO("%s", parse_info.c_str());
 
+	start_web_server();
+
 	if (printme) {
 #ifdef __APPLE_
 		fl_open_display();
 #endif
 		print_and_exit();
-		if (exit_after_print)
+		if (exit_after_print && selected_form != CUSTOM)
 			return 0;
 	}
 
@@ -2070,7 +2120,11 @@ int main(int argc, char *argv[])
 
 	Fl::add_timeout(0.10, after_start);
 
-	return Fl::run();
+	int result = Fl::run();
+
+	if (server) mg_destroy_server(&server);
+
+	return result;
 }
 
 void print_and_exit()
@@ -2148,6 +2202,10 @@ void print_and_exit()
 			cb_csv_save();
 			cb_csv_html();
 			break;
+		case CUSTOM :
+			cb_custom_save();
+			cb_custom_html();
+			break;
 		case MARSDAILY :
 			cb_mars_daily_save();
 			cb_mars_daily_html();
@@ -2265,9 +2323,9 @@ void checkdirectories(void)
 		{ ICS_msg_dir,    "ICS/messages", 0 },
 		{ ICS_tmp_dir,    "ICS/templates", 0 },
 		{ CSV_dir,        "CSV", 0},
+		{ CUSTOM_dir,     "CUSTOM", 0},
 		{ XFR_dir,        "TRANSFERS", 0},
-		{ FLMSG_temp_dir, "temp_files", 0 },
-		{ FLMSG_custom_dir, "CUSTOM", 0 }
+		{ FLMSG_temp_dir, "temp_files", 0 }
 	};
 
 	int r;
@@ -2365,6 +2423,11 @@ void show_help()
 	open_url("http://www.w1hkj.com/flmsg-help/index.html");
 }
 
+void custom_download()
+{
+	open_url("https://groups.yahoo.com/neo/groups/NBEMSham/files/Custom_flmsg_forms/");
+}
+
 int parse_args(int argc, char **argv, int& idx)
 {
 	if (strstr(argv[idx], "--p")) {
@@ -2468,6 +2531,9 @@ int parse_args(int argc, char **argv, int& idx)
 		fname.find(CSVFILE_EXT) != string::npos ||
 		fname.find(CSVTEMP_EXT) != string::npos ||
 
+		fname.find(CUSTOMFILE_EXT) != string::npos ||
+		fname.find(CUSTOMTEMP_EXT) != string::npos ||
+
 		fname.find(FMARSDAILY_EXT) != string::npos ||
 		fname.find(TMARSDAILY_EXT) != string::npos ||
 
@@ -2557,3 +2623,177 @@ LOG_INFO("%s", url);
 		fl_alert2(_("Could not open url:\n%s\n"), url);
 #endif
 }
+
+//------------------------------------------------------------------------------
+// mongoose server support
+//------------------------------------------------------------------------------
+bool exit_server = false;
+pthread_t *web_server_thread = 0;
+pthread_mutex_t mutex_web_server = PTHREAD_MUTEX_INITIALIZER;
+
+int handle_type = HANDLE_WAITING;
+
+void remove_port_number()
+{
+	string ports;
+	string chk_fname = FLMSG_dir;
+	chk_fname.append("port_names.txt");
+	ifstream rchkfile(chk_fname.c_str());
+	if (rchkfile) {
+		string pnbr;
+		rchkfile >> pnbr;
+		while (!rchkfile.eof()) {
+			ports.append(pnbr).append("\n");
+			rchkfile >> pnbr;
+		}
+		rchkfile.close();
+		size_t p = ports.find(flmsg_webserver_szportnbr);
+		if (p != string::npos) ports.erase(p, 5);
+		ofstream wchkfile(chk_fname.c_str());
+		if (wchkfile) {
+			wchkfile << ports;
+			wchkfile.close();
+		}
+	}
+	LOG_INFO("Closed Web server on localhost:%s", flmsg_webserver_szportnbr);
+}
+
+void get_next_port_number()
+{
+	string ports;
+	string chk_fname = FLMSG_dir;
+	chk_fname.append("port_names.txt");
+	ifstream rchkfile(chk_fname.c_str());
+	if (rchkfile) {
+		string pnbr;
+		rchkfile >> pnbr;
+		while (!rchkfile.eof()) {
+			ports.append(pnbr).append("\n");
+			rchkfile >> pnbr;
+		}
+		rchkfile.close();
+		if (ports.empty()) {
+			ports = "8080\n";
+			flmsg_webserver_portnbr = 8080;
+			strcpy(flmsg_webserver_szportnbr, "8080");
+		} else {
+			flmsg_webserver_portnbr = 8080;
+			snprintf(flmsg_webserver_szportnbr, 
+					sizeof(flmsg_webserver_szportnbr),
+					"%d", flmsg_webserver_portnbr);
+			while (ports.find(flmsg_webserver_szportnbr) != string::npos) {
+				flmsg_webserver_portnbr++;
+				snprintf(flmsg_webserver_szportnbr, 
+						sizeof(flmsg_webserver_szportnbr),
+						"%d", flmsg_webserver_portnbr);
+			}
+			ports.append(flmsg_webserver_szportnbr).append("\n");
+		}
+		ofstream wchkfile(chk_fname.c_str());
+		if (wchkfile) {
+			wchkfile << ports;
+			wchkfile.close();
+		}
+	} else {
+		ports = "8080\n";
+		flmsg_webserver_portnbr = 8080;
+		strcpy(flmsg_webserver_szportnbr, "8080");
+		ofstream wchkfile(chk_fname.c_str());
+		if (wchkfile) {
+			wchkfile << ports;
+			wchkfile.close();
+		}
+	}
+}
+
+
+static const char *html_waiting =
+  "<html><body>\n\
+	Custom form not posted<br>\n\
+	</html>";
+
+void load_custom(void *)
+{
+	load_custom_menu();
+}
+
+void * poll_server(void *d)
+{
+	int n = 10;
+	exit_server = false;
+
+	while(!exit_server) {
+		pthread_mutex_lock(&mutex_web_server);
+			mg_poll_server(server, 200);
+		pthread_mutex_unlock(&mutex_web_server);
+		n--;
+		if (!n) {
+			Fl::awake(load_custom, 0);
+			n = 10;
+		}
+		MilliSleep(100);
+	}
+	return NULL;
+}
+
+void close_server()
+{
+	pthread_mutex_lock(&mutex_web_server);
+	exit_server = true;
+	pthread_mutex_unlock(&mutex_web_server);
+	remove_port_number();
+	MilliSleep(200);
+}
+
+extern void get_html_vars(struct mg_connection *conn);
+static int web_handler(struct mg_connection *conn)
+{
+	if (strcmp(conn->uri, "/handle_post_request") == 0) {
+		get_html_vars(conn);
+		custom_viewer(conn);
+  } else {
+		if (handle_type == HANDLE_EDIT)
+			custom_editor(conn);
+		else if (handle_type == HANDLE_VIEW)
+			custom_viewer(conn);
+		else if (handle_type == HANDLE_WAITING)
+// Show HTML waiting
+			mg_send_data(conn, html_waiting, strlen(html_waiting));
+	}
+	handle_type = HANDLE_NONE;
+
+	return MG_REQUEST_PROCESSED;
+}
+
+void start_web_server()
+{
+	if ((server = mg_create_server(NULL)) == NULL) {
+		fl_alert2("%s", "Failed to start web server");
+		return;
+	}
+
+	const char *msg = mg_set_option(server, "document_root", CUSTOM_dir.c_str());
+
+	if (msg != NULL) {
+		fl_alert2("%s", "Failed to set file server root directory");
+		return;
+	}
+
+	msg = mg_set_option(server, "listening_port", flmsg_webserver_szportnbr);
+	if (msg) {
+		LOG_ERROR("%s %s", msg, flmsg_webserver_szportnbr);
+		fl_alert2("Failed to open web server port %s", flmsg_webserver_szportnbr);
+		return;
+	}
+
+	mg_set_request_handler(server, web_handler);
+
+	web_server_thread = new pthread_t;
+	if (pthread_create(web_server_thread, NULL, poll_server, NULL)) {
+		perror("pthread_create");
+		exit(EXIT_FAILURE);
+	}
+
+	LOG_INFO("Web server on localhost:%s", flmsg_webserver_szportnbr);
+
+}
